# File: settings.py
# File Created: Thursday, 11th July 2019 12:13:30 am
# Author: Steven Atkinson (steven@atkinson.mn)

from torch.distributions.transforms import ExpTransform

DefaultPositiveTransform = ExpTransform
